from typing import Optional

import jax.numpy as jnp
from chex import dataclass
from jax import vmap
from multipledispatch import dispatch

from .types import Array


@dataclass
class Kernel:
    ndims: Optional[int] = 1
    stationary: Optional[bool] = False
    spectral: Optional[bool] = False
    name: Optional[str] = "Kernel"

    def __call__(self, x: Array, y: Array) -> Array:
        raise NotImplementedError

    @property
    def ard(self):
        return True if self.ndims > 1 else False


@dataclass
class RBF(Kernel):
    ndims = 1
    stationary = True
    spectral = False
    name = "RBF"

    def __call__(self, x: Array, y: Array) -> Array:
        return jnp.exp(-0.5 * squared_distance(x, y))


@dispatch(RBF)
def initialise(kernel: RBF):
    return {"lengthscale": jnp.array([1.0]*kernel.ndims), "variance": jnp.array([1.0])}


def squared_distance(x: Array, y: Array):
    return jnp.sum((x - y) ** 2)


def gram(kernel: Kernel, inputs: Array) -> Array:
    return vmap(lambda x1: vmap(lambda y1: kernel(x1, y1))(inputs))(inputs)


def cross_covariance(kernel: Kernel, x: Array, y: Array) -> Array:
    return vmap(lambda x1: vmap(lambda y1: kernel(x1, y1))(x))(y)



import jax.numpy as jnp
from multipledispatch import dispatch

from ..gps import ConjugatePosterior, NonConjugatePosterior, Posterior
from ..kernels import Kernel, initialise
from ..likelihoods import Likelihood, initialise
from ..mean_functions import MeanFunction, initialise
from ..utils import concat_dictionaries, merge_dictionaries


def _initialise_hyperparams(kernel: Kernel, meanf: MeanFunction) -> dict:
    return concat_dictionaries(initialise(kernel), initialise(meanf))


@dispatch(ConjugatePosterior)
def initialise(gp: ConjugatePosterior) -> dict:
    hyps = _initialise_hyperparams(gp.prior.kernel, gp.prior.mean_function)
    return concat_dictionaries(hyps, initialise(gp.likelihood))


@dispatch(ConjugatePosterior, object)
def initialise(gp: ConjugatePosterior, n_data):
    return initialise(gp)


@dispatch(NonConjugatePosterior, int)
def initialise(gp: NonConjugatePosterior, n_data: int) -> dict:
    hyperparams = _initialise_hyperparams(gp.prior.kernel, gp.prior.mean_function)
    likelihood = concat_dictionaries(hyperparams, initialise(gp.likelihood))
    latent_process = {"latent": jnp.zeros(shape=(n_data, 1))}
    return concat_dictionaries(likelihood, latent_process)


def complete(params: dict, gp: Posterior, n_data: int = None) -> dict:
    full_param_set = initialise(gp, n_data)
    return merge_dictionaries(full_param_set, params)



import jax.numpy as jnp
from jax.interpreters.ad import JVPTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd
from ..gps import NonConjugatePosterior
import warnings

from ..types import Array, NoneType


@dispatch((jnp.DeviceArray, JVPTracer, DynamicJaxprTracer), tfd.Distribution)
def log_density(param: jnp.DeviceArray, density: tfd.Distribution) -> Array:
    return density.log_prob(param)


@dispatch(dict, NoneType)
def evaluate_prior(params: dict, priors: dict) -> Array:
    return jnp.array(0.)


@dispatch(dict, dict)
def evaluate_prior(params: dict, priors: dict) -> Array:
    lpd = jnp.array(0)
    for param, val in priors.items():
        lpd+=jnp.sum(log_density(params[param], priors[param]))
    return lpd


@dispatch(NonConjugatePosterior, dict)
def prior_checks(gp: NonConjugatePosterior, priors: dict) -> dict:
    if 'latent' in priors.keys():
        latent_prior = priors['latent']
        if latent_prior.name != 'Normal':
            warnings.warn(f'A {latent_prior.name} distribution prior has been placed on the latent function. It is strongly afvised that a unit-Gaussian prior is used.')
        return priors
    else:
        priors['latent'] = tfd.Normal(loc=0., scale=1.)
        return priors


from typing import Tuple, Union

import jax.numpy as jnp
from chex import Array
from jax.interpreters.pxla import ShardedDeviceArray

# Array = Union[jnp.ndarray, ShardedDeviceArray, jnp.DeviceArray]  # Cannot currently dispatch on a Union type
# Data = Tuple[Array, Array]
NoneType = type(None)


from typing import Callable, List

import jax.numpy as jnp
from chex import dataclass

from ..types import Array


def softplus(x: Array) -> Array:
    """
    Map an input value x from a constrained space into an unconstrained space using the softplus transformation
    :math:`\log(\exp(x)-1)`.

    :param x: A constrained value that exists on the positive real line
    :return: An unconstrained representation of x that exists on the entire real line
    """
    return jnp.log(jnp.exp(x) - 1.0)


def inverse_softplus(x: Array) -> Array:
    """
    Reverse the softplus transformation such that the unconstrained representation now exists on the positive real
    line again using  :math:`\log(1+\exp(x))`.

    :param x: An unconstrained representation of x that exists on the entire real line
    :return: A constrained value that exists on the positive real line
    """
    return jnp.log(1.0 + jnp.exp(x))


def identity(x: Array) -> Array:
    return x


@dataclass
class Transformation:
    forward: Callable
    backward: Callable


@dataclass
class SoftplusTransformation(Transformation):
    forward: Callable = softplus
    backward: Callable = inverse_softplus


@dataclass
class IdentityTransformation(Transformation):
    forward: Callable = identity
    backward: Callable = identity


def transform(params: dict, transformation: Transformation):
    transformed_params = {}
    for k, v in params.items():
        if k != "latent":
            transformed_params[k] = transformation.forward(v)
        else:
            transformed_params[k] = v
    return transformed_params


def untransform(params: dict, transformation: Transformation):
    untransformed_params = {}
    for k, v in params.items():
        if k != "latent":
            untransformed_params[k] = transformation.backward(v)
        else:
            untransformed_params[k] = v
    return untransformed_params



import jax.numpy as jnp
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd

from .gps import ConjugatePosterior, Prior
from .kernels import gram
from .predict import mean, variance
from .types import Array
from .utils import I


@dispatch(Prior, dict, jnp.DeviceArray)
def random_variable(
    gp: Prior, params: dict, sample_points: Array, jitter_amount: float = 1e-6
) -> tfd.Distribution:
    mu = gp.mean_function(sample_points)
    gram_matrix = params["variance"] * gram(gp.kernel, sample_points / params["lengthscale"])
    jitter_matrix = I(sample_points.shape[0]) * jitter_amount
    covariance = gram_matrix + jitter_matrix
    return tfd.MultivariateNormalFullCovariance(mu.squeeze(), covariance)


@dispatch(ConjugatePosterior, dict, jnp.DeviceArray, jnp.DeviceArray, jnp.DeviceArray)
def random_variable(
    gp: ConjugatePosterior,
    params: dict,
    sample_points: Array,
    train_inputs: Array,
    train_outputs: Array,
    jitter_amount: float = 1e-6,
) -> tfd.Distribution:
    n = sample_points.shape[0]
    # TODO: Return kernel matrices here to avoid replicated computation.
    mu = mean(gp, params, sample_points, train_inputs, train_outputs)
    cov = variance(gp, params, sample_points, train_inputs, train_outputs)
    return tfd.MultivariateNormalFullCovariance(mu.squeeze(), cov + I(n) * jitter_amount)


@dispatch(jnp.DeviceArray, Prior, dict, jnp.DeviceArray)
def sample(key, gp, params, sample_points, n_samples=1) -> Array:
    rv = random_variable(gp, params, sample_points)
    return rv.sample(sample_shape=(n_samples,), seed=key)


@dispatch(jnp.DeviceArray, tfd.Distribution)
def sample(key: jnp.DeviceArray, random_variable: tfd.Distribution, n_samples: int = 1) -> Array:
    return random_variable.sample(sample_shape=(n_samples,), seed=key)



import jax.numpy as jnp
from jax.scipy.linalg import cho_factor, cho_solve, cholesky, solve_triangular
from multipledispatch import dispatch

from .gps import ConjugatePosterior, NonConjugatePosterior, Prior
from .kernels import cross_covariance, gram
from .likelihoods import predictive_moments
from .types import Array
from .utils import I


@dispatch(ConjugatePosterior, dict, jnp.DeviceArray, jnp.DeviceArray, jnp.DeviceArray)
def mean(
    gp: ConjugatePosterior,
    param: dict,
    test_inputs: Array,
    train_inputs: Array,
    train_outputs: Array,
):
    assert (
        train_outputs.ndim == 2
    ), f"2-dimensional training outputs are required. Current dimensional: {train_outputs.ndim}."
    ell, alpha = param["lengthscale"], param["variance"]
    sigma = param["obs_noise"]
    n_train = train_inputs.shape[0]

    Kff = alpha * gram(gp.prior.kernel, train_inputs / ell)
    Kfx = alpha * cross_covariance(gp.prior.kernel, train_inputs / ell, test_inputs / ell)

    prior_mean = gp.prior.mean_function(train_inputs)
    L = cho_factor(Kff + I(n_train) * sigma, lower=True)

    prior_distance = train_outputs - prior_mean
    weights = cho_solve(L, prior_distance)
    return jnp.dot(Kfx, weights)


@dispatch(ConjugatePosterior, dict, jnp.DeviceArray, jnp.DeviceArray, jnp.DeviceArray)
def variance(
    gp: ConjugatePosterior,
    param: dict,
    test_inputs: Array,
    train_inputs: Array,
    train_outputs: Array,
) -> Array:
    assert (
        train_outputs.ndim == 2
    ), f"2-dimensional training outputs are required. Current dimensional: {train_outputs.ndim}."
    ell, alpha = param["lengthscale"], param["variance"]
    sigma = param["obs_noise"]
    n_train = train_inputs.shape[0]
    n_test = test_inputs.shape[0]

    Kff = alpha * gram(gp.prior.kernel, train_inputs / ell)
    Kfx = alpha * cross_covariance(gp.prior.kernel, train_inputs / ell, test_inputs / ell)
    Kxx = alpha * gram(gp.prior.kernel, test_inputs / ell)

    L = cho_factor(Kff + I(n_train) * sigma, lower=True)
    latents = cho_solve(L, Kfx.T)
    return Kxx - jnp.dot(Kfx, latents)


@dispatch(
    NonConjugatePosterior,
    dict,
    jnp.DeviceArray,
    jnp.DeviceArray,
    jnp.DeviceArray,
)
def mean(
    gp: NonConjugatePosterior,
    param: dict,
    test_inputs: Array,
    train_inputs: Array,
    train_outputs: Array,
):
    ell, alpha, nu = param["lengthscale"], param["variance"], param["latent"]
    n_train = train_inputs.shape[0]
    Kff = alpha * gram(gp.prior.kernel, train_inputs / ell)
    Kfx = alpha * cross_covariance(gp.prior.kernel, train_inputs / ell, test_inputs / ell)
    Kxx = alpha * gram(gp.prior.kernel, test_inputs / ell)
    L = jnp.linalg.cholesky(Kff + jnp.eye(train_inputs.shape[0]) * 1e-6)

    A = solve_triangular(L, Kfx.T, lower=True)
    latent_var = Kxx - jnp.sum(jnp.square(A), -2)
    latent_mean = jnp.matmul(A.T, nu)

    lvar = jnp.diag(latent_var)

    moment_fn = predictive_moments(gp.likelihood)
    pred_rv = moment_fn(latent_mean.ravel(), lvar)
    return pred_rv.mean()


@dispatch(
    NonConjugatePosterior,
    dict,
    jnp.DeviceArray,
    jnp.DeviceArray,
    jnp.DeviceArray,
)
def variance(
    gp: NonConjugatePosterior,
    param: dict,
    test_inputs: Array,
    train_inputs: Array,
    train_outputs: Array,
):
    ell, alpha, nu = param["lengthscale"], param["variance"], param["latent"]
    n_train = train_inputs.shape[0]
    Kff = alpha * gram(gp.prior.kernel, train_inputs / ell)
    Kfx = alpha * cross_covariance(gp.prior.kernel, train_inputs / ell, test_inputs / ell)
    Kxx = alpha * gram(gp.prior.kernel, test_inputs / ell)
    L = jnp.linalg.cholesky(Kff + jnp.eye(train_inputs.shape[0]) * 1e-6)

    A = solve_triangular(L, Kfx.T, lower=True)
    latent_var = Kxx - jnp.sum(jnp.square(A), -2)
    latent_mean = jnp.matmul(A.T, nu)
    lvar = jnp.diag(latent_var)
    moment_fn = predictive_moments(gp.likelihood)
    pred_rv = moment_fn(latent_mean.ravel(), lvar)
    return pred_rv.variance()



from typing import Optional

import jax.numpy as jnp
from chex import dataclass
from multipledispatch import dispatch

from .types import Array


@dataclass
class MeanFunction:
    output_dim: Optional[int] = 1
    name: Optional[str] = "Mean Function"

    def __call__(self, x: Array) -> Array:
        raise NotImplementedError


@dataclass
class Zero(MeanFunction):
    output_dim: Optional[int] = 1
    name: Optional[str] = "Zero mean"

    def __call__(self, x: Array) -> Array:
        out_shape = (x.shape[0], self.output_dim)
        return jnp.zeros(shape = out_shape)


@dispatch(Zero)
def initialise(meanf: Zero) -> dict:
    return {}
